"""Generate Golang code for enhancing model classes."""

import io
from typing import Tuple, Optional, List, Sequence

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_wrap_forward_declarations(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the forward declarations of the overloaded ``Wrap`` functions."""
    result = []  # type: List[Stripped]
    for cls in symbol_table.classes:
        interface_name = cpp_naming.interface_name(cls.name)

        function_name = cpp_naming.function_name(Identifier("wrap"))

        result.append(
            Stripped(
                f"""\
template <typename E>
std::shared_ptr<
{I}types::{interface_name}
> {function_name}(
{I}const std::shared_ptr<
{II}types::{interface_name}
{I}>& that,
{I}const std::function<
{II}std::shared_ptr<E>(
{III}const std::shared_ptr<types::IClass>&
{II})
{I}>& factory
);"""
            )
        )

    return result


def _generate_enhanced_interface_definition() -> Stripped:
    """Generate the definition of the generic ``Enhanced`` interface."""
    enhanced_interface = cpp_naming.interface_name(Identifier("Enhanced"))

    members = []  # type: List[Stripped]

    enhancement_getter = cpp_naming.getter_name(Identifier("enhancement"))
    members.append(
        Stripped(
            f"""\
virtual const std::shared_ptr<E>& {enhancement_getter}() const = 0;"""
        )
    )

    enhancement_mutable_getter = cpp_naming.mutable_getter_name(
        Identifier("enhancement")
    )
    members.append(
        Stripped(
            f"""\
virtual std::shared_ptr<E>& {enhancement_mutable_getter}() = 0;"""
        )
    )

    enhancement_setter = cpp_naming.setter_name(Identifier("enhancement"))
    members.append(
        Stripped(
            f"""\
virtual void {enhancement_setter}(
{I}std::shared_ptr<E> value
) = 0;"""
        )
    )

    members.append(
        Stripped(
            f"""\
virtual ~{enhanced_interface}() = default;"""
        )
    )

    members_joined = "\n\n".join(members)

    return Stripped(
        f"""\
template<typename E>
class {enhanced_interface} {{
 public:
{I}{indent_but_first_line(members_joined, I)}
}};"""
    )


def _generate_method_delegation(method: intermediate.Method) -> Stripped:
    """Generate the delegated method to ``instance_``."""
    returns = (
        cpp_common.generate_type(method.returns, types_namespace=Identifier("types"))
        if method.returns is not None
        else None
    )

    return_type = "void" if returns is None else returns
    return_prefix = "" if returns is None else "return "

    arg_types_names = [
        (
            cpp_common.generate_type_with_const_ref_if_applicable(
                arg.type_annotation, types_namespace=Identifier("types")
            ),
            cpp_naming.argument_name(arg.name),
        )
        for arg in method.arguments
    ]

    method_name = cpp_naming.method_name(method.name)

    const_suffix = " const" if method.non_mutating else ""

    if len(method.arguments) == 0:
        return Stripped(
            f"""\
{return_type} {method_name}(){const_suffix} override {{
{I}{return_prefix}instance_->{method_name}();
}}"""
        )

    arguments_definition = ",\n".join(
        f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
    )

    arguments_delegation = ",\n".join(f"{arg_name}" for _, arg_name in arg_types_names)

    return Stripped(
        f"""\
{return_type} {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
){const_suffix} override {{
{I}{return_prefix}instance_->{method_name}(
{II}{indent_but_first_line(arguments_delegation, II)}
{I});
}}"""
    )


def _generate_enhanced_class(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the implementation of an enhanced class which wraps an instance."""
    public_members = []  # type: List[Stripped]

    model_type_enum = cpp_naming.enum_name(Identifier("Model_type"))
    model_type_getter = cpp_naming.getter_name(Identifier("model_type"))

    model_type_literal = cpp_naming.enum_literal_name(Identifier(cls.name))

    public_members.append(
        Stripped(
            f"""\
types::{model_type_enum} {model_type_getter}() const override {{
{I}return types::{model_type_enum}::{model_type_literal};
}}"""
        )
    )

    for prop in cls.properties:
        getter_name = cpp_naming.getter_name(prop.name)

        getter_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=prop.type_annotation, types_namespace=Identifier("types")
        )

        public_members.append(
            Stripped(
                f"""\
{getter_type} {getter_name}() const override {{
{I}return instance_->{getter_name}();
}}"""
            )
        )

        mutable_getter_name = cpp_naming.mutable_getter_name(prop.name)
        mutable_getter_type = cpp_common.generate_type_with_ref(
            type_annotation=prop.type_annotation, types_namespace=Identifier("types")
        )
        public_members.append(
            Stripped(
                f"""\
{mutable_getter_type} {mutable_getter_name}() override {{
{I}return instance_->{mutable_getter_name}();
}}"""
            )
        )

        setter_name = cpp_naming.setter_name(prop.name)

        # NOTE (mristin, 2023-07-05):
        # For a discussion on ``std::shared_ptr`` and referencing,
        # see: https://herbsutter.com/2013/06/05/gotw-91-solution-smart-pointer-parameters/

        # NOTE (mristin, 2023-07-07):
        # We provide setters which only set-by-value logic since this seems to be
        # the best approach for general cases,
        # see: https://stackoverflow.com/questions/10692345/is-it-worth-adding-a-move-enabled-setter.
        #
        # Whenever you know that you do not need the value after calling the setter,
        # make sure to call it with ``std::move(.)``.

        value_type = cpp_common.generate_type(
            type_annotation=prop.type_annotation, types_namespace=Identifier("types")
        )
        public_members.append(
            Stripped(
                f"""\
void {setter_name}(
{I}{indent_but_first_line(value_type, I)} value
) override {{
{I}instance_->{setter_name}(value);
}}"""
            )
        )

    for method in cls.methods:
        public_members.append(_generate_method_delegation(method=method))

    enhancement_getter = cpp_naming.getter_name(Identifier("enhancement"))
    public_members.append(
        Stripped(
            f"""\
const std::shared_ptr<E>& {enhancement_getter}() const {{
{I}return enhancement_;
}}"""
        )
    )

    enhancement_mutable_getter = cpp_naming.mutable_getter_name(
        Identifier("enhancement")
    )
    public_members.append(
        Stripped(
            f"""\
std::shared_ptr<E>& {enhancement_mutable_getter}() {{
{I}return enhancement_;
}}"""
        )
    )

    # NOTE (mristin, 2023-07-07):
    # See: https://stackoverflow.com/questions/41871115/why-would-i-stdmove-an-stdshared-ptr
    # for why we ``std::move`` here.
    enhancement_setter = cpp_naming.setter_name(Identifier("enhancement"))
    public_members.append(
        Stripped(
            f"""\
void {enhancement_setter}(
{I}std::shared_ptr<E> value
) {{
{I}enhancement_ = std::move(value);
}}"""
        )
    )

    enhanced_cls_name = cpp_naming.class_name(Identifier(f"enhanced_{cls.name}"))
    interface_name = cpp_naming.interface_name(cls.name)

    public_members.append(
        Stripped(
            f"""\
{enhanced_cls_name}(
{I}std::shared_ptr<types::{interface_name}> instance,
{I}std::shared_ptr<E> enhancement
) :
{I}instance_(instance),
{I}enhancement_(enhancement) {{
{I}// Intentionally empty.
}}"""
        )
    )

    public_members.append(
        Stripped(
            f"""\
virtual ~{enhanced_cls_name}() = default;"""
        )
    )

    public_members_joined = "\n\n".join(public_members)
    enhanced_interface = cpp_naming.interface_name(Identifier("Enhanced"))

    return Stripped(
        f"""\
template<class E>
class {enhanced_cls_name}
{II}: virtual public types::{interface_name},
{II}virtual public {enhanced_interface}<E> {{
 public:
{I}{indent_but_first_line(public_members_joined, I)}

 private:
{I}std::shared_ptr<types::{interface_name}> instance_;
{I}std::shared_ptr<E> enhancement_;
}};"""
    )


# NOTE (mristin, 2023-07-07):
# We write two separate functions, ``_generate_wrap_snippet_for_required_property`` and
# ``_generate_wrap_snippet_for_optional_property``, as the complexity grew over the top.
# This resulted in a much more readable code than if we tried to de-DRY the logic
# in a single function.


@require(lambda prop: not isinstance(prop, intermediate.OptionalTypeAnnotation))
def _generate_wrap_snippet_for_required_property(
    prop: intermediate.Property,
) -> Stripped:
    """
    Generate the snippet to recursively wrap the required property.

    We return an empty string if there is no snippet for the property.
    """
    type_anno = prop.type_annotation

    # NOTE (mristin, 2023-07-07):
    # Duplicate the pre-condition for mypy.
    assert not isinstance(type_anno, intermediate.OptionalTypeAnnotation)

    setter_name = cpp_naming.setter_name(prop.name)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # Nothing to recurse into.
        return Stripped("")

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            # Nothing to recurse into.
            return Stripped("")

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            # Nothing to recurse into.
            return Stripped("")

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-07-07):
            # The non-mutating getter means here that we will not change the reference,
            # but we want to recurse into the object.
            getter_name = cpp_naming.getter_name(prop.name)

            return Stripped(
                f"""\
that->{setter_name}(
{I}Wrap<E>(
{II}that->{getter_name}(),
{II}factory
{I})
);"""
            )
        else:
            assert_never(type_anno.our_type)
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-07-07): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        getter_name = cpp_naming.getter_name(prop.name)
        const_ref_prop_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=type_anno, types_namespace=Identifier("types")
        )
        prop_type = cpp_common.generate_type(
            type_annotation=type_anno, types_namespace=cpp_common.TYPES_NAMESPACE
        )

        item_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=type_anno.items, types_namespace=cpp_common.TYPES_NAMESPACE
        )

        return Stripped(
            f"""\
{{
{I}{indent_but_first_line(const_ref_prop_type, I)} value(
{II}that->{getter_name}()
{I});
{I}const std::size_t size = value.size();

{I}{indent_but_first_line(prop_type, I)} wrapped;
{I}wrapped.reserve(size);

{I}for (
{II}{indent_but_first_line(item_type, II)} item
{II}: value
{I}) {{
{II}wrapped.emplace_back(
{III}Wrap<E>(
{IIII}item,
{IIII}factory
{III})
{II});
{I}}}

{I}that->{setter_name}(
{II}std::move(wrapped)
{I});
}}"""
        )
    else:
        assert_never(type_anno)


# fmt: off
@require(
    lambda prop:
    isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
)
# fmt: on
def _generate_wrap_snippet_for_optional_property(
    prop: intermediate.Property,
) -> Stripped:
    """
    Generate the snippet to recursively wrap the optional property.

    We return an empty string if there is no snippet for the property.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    setter_name = cpp_naming.setter_name(prop.name)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # Nothing to recurse into.
        return Stripped("")

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            # Nothing to recurse into.
            return Stripped("")

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            # Nothing to recurse into.
            return Stripped("")

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-07-07):
            # The non-mutating getter means here that we will not change the reference,
            # but we want to recurse into the object.
            getter_name = cpp_naming.getter_name(prop.name)

            value_type = cpp_common.generate_type(
                type_annotation=type_anno, types_namespace=Identifier("types")
            )

            value_interface_name = cpp_naming.interface_name(type_anno.our_type.name)

            return Stripped(
                f"""\
if (that->{getter_name}().has_value()) {{
{I}const {indent_but_first_line(value_type, II)}& value(
{II}that->{getter_name}().value()
{I});

{I}std::shared_ptr<
{II}types::{value_interface_name}
{I}> wrapped(
{II}Wrap<E>(
{III}value,
{III}factory
{II})
{I});

{I}that->{setter_name}(
{II}common::make_optional(
{III}std::move(wrapped)
{II})
{I});
}}"""
            )
        else:
            assert_never(type_anno.our_type)
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-07-07): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        getter_name = cpp_naming.getter_name(prop.name)
        value_type = cpp_common.generate_type(
            type_annotation=type_anno, types_namespace=cpp_common.TYPES_NAMESPACE
        )

        item_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=type_anno.items, types_namespace=cpp_common.TYPES_NAMESPACE
        )

        return Stripped(
            f"""\
if (that->{getter_name}().has_value()) {{
{I}const {indent_but_first_line(value_type, II)}& value(
{II}that->{getter_name}().value()
{I});
{I}const std::size_t size = value.size();

{I}{indent_but_first_line(value_type, I)} wrapped;
{I}wrapped.reserve(size);

{I}for (
{II}{indent_but_first_line(item_type, II)} item
{II}: value
{I}) {{
{II}wrapped.emplace_back(
{III}Wrap<E>(
{IIII}item,
{IIII}factory
{III})
{II});
{I}}}

{I}that->{setter_name}(
{II}common::make_optional(
{III}std::move(wrapped)
{II})
{I});
}}"""
        )
    else:
        assert_never(type_anno)


def _generate_concrete_wrap(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the concrete wrapping function for the concrete class."""
    recurse_blocks = []  # type: List[Stripped]
    for prop in cls.properties:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            recurse_block = _generate_wrap_snippet_for_optional_property(
                prop=prop,
            )
        else:
            recurse_block = _generate_wrap_snippet_for_required_property(
                prop=prop,
            )

        if len(recurse_block) != 0:
            recurse_blocks.append(recurse_block)

    recurse_blocks_joined = Stripped(
        "\n\n".join(recurse_blocks)
        if len(recurse_blocks) > 0
        else "// No properties to be recursively enhanced."
    )

    interface_name = cpp_naming.interface_name(cls.name)
    enhanced_cls_name = cpp_naming.class_name(Identifier(f"enhanced_{cls.name}"))

    blocks = [
        Stripped(
            """\
// We assume that we already checked whether `that` has been enhanced
// in the caller."""
        ),
        recurse_blocks_joined,
        Stripped(
            f"""\
std::shared_ptr<E> enh(
{I}factory(that)
);
return (enh == nullptr)
{I}? that
{I}: std::shared_ptr<types::{interface_name}>(
{II}new {enhanced_cls_name}<E>(
{III}that,
{III}enh
{II})
{I});"""
        ),
    ]

    body = "\n\n".join(blocks)

    function_name = cpp_naming.function_name(Identifier(f"wrap_{cls.name}"))

    return Stripped(
        f"""\
/**
 * Wrap \\p that with an enhanced instance.
 *
 * \\param that instance to be wrapped and enhanced
 * \\param factory to produce an enhancement based on an instance
 * \\return Enhanced instance, or `that` if no enhancement produced
 *
 * \\tparam E type of the enhancement
 */
template<typename E>
std::shared_ptr<types::{interface_name}> {function_name}(
{I}const std::shared_ptr<types::{interface_name}>& that,
{I}const std::function<
{II}std::shared_ptr<E>(
{III}const std::shared_ptr<types::IClass>&
{II})
{I}>& factory
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_wrap_for(
    interface_name: Identifier, concrete_classes: Sequence[intermediate.ConcreteClass]
) -> Stripped:
    """
    Generate the wrap function for the given interface.

    We do not operate on ``intermediate.Class`` directly as we also want to
    handle the case for the most abstract ``IClass``.
    """
    model_type_name = cpp_naming.enum_name(Identifier("Model_type"))

    case_blocks = []  # type: List[Stripped]
    for cls in concrete_classes:
        literal = cpp_naming.enum_literal_name(literal_name=cls.name)

        concrete_wrap_function = cpp_naming.function_name(
            Identifier(f"wrap_{cls.name}")
        )

        concrete_interface_name = cpp_naming.interface_name(cls.name)

        if concrete_interface_name == interface_name:
            case_blocks.append(
                Stripped(
                    f"""\
case types::{model_type_name}::{literal}:
{I}return impl::{concrete_wrap_function}<E>(
{II}that,
{II}factory
{I});
{I}break;"""
                )
            )
        else:
            case_blocks.append(
                Stripped(
                    f"""\
case types::{model_type_name}::{literal}:
{I}return impl::{concrete_wrap_function}<E>(
{II}std::dynamic_pointer_cast<
{III}types::{concrete_interface_name}
{II}>(that),
{II}factory
{I});
{I}break;"""
                )
            )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected model type: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(
{IIIII}that->model_type()
{IIII})
{III})
{II})
{I});
{I}break;"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    # noinspection SpellCheckingInspection
    body = Stripped(
        f"""\
impl::AssertNotEnhanced<
{I}E,
{I}types::{interface_name}
>(that);

switch (that->model_type()) {{
{I}{indent_but_first_line(case_blocks_joined, I)}
}}"""
    )

    function_name = cpp_naming.function_name(Identifier("wrap"))

    return Stripped(
        f"""\
template <typename E>
std::shared_ptr<
{I}types::{interface_name}
> {function_name}(
{I}const std::shared_ptr<
{II}types::{interface_name}
{I}>& that,
{I}const std::function<
{II}std::shared_ptr<E>(
{III}const std::shared_ptr<types::IClass>&
{II})
{I}>& factory
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_wrap(symbol_table: intermediate.SymbolTable) -> List[Stripped]:
    """Generate all the overloads of the main wrapping function."""
    blocks = [
        _generate_wrap_for(
            interface_name=Identifier("IClass"),
            concrete_classes=symbol_table.concrete_classes,
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.classes:
        interface_name = cpp_naming.interface_name(cls.name)

        concrete_classes = []  # type: List[intermediate.ConcreteClass]
        if isinstance(cls, intermediate.ConcreteClass):
            concrete_classes.append(cls)

        concrete_classes.extend(cls.concrete_descendants)

        blocks.append(
            _generate_wrap_for(
                interface_name=interface_name, concrete_classes=concrete_classes
            )
        )

    return [
        Stripped(
            """\
/**
 * Wrap \\p that instance recursively with the enhancement produced by the \\p factory.
 *
 * The factory decides itself whether it will produce an enhancement for
 * \\p that instance, or not. Even if no enhancement has been produced for \\p that
 * instance, we will still continue to enhance the instances referenced
 * by \\p that instance recursively.
 *
 * \\param that instance to wrap
 * \\param factory to selectively produce an enhancement
 * \\return enhanced \\p that instance
 * \\throw std::logic_error if \\p that instance has been already wrapped.
 * \\tparam E type of the enhancement
 */
///@{"""
        ),
        *blocks,
        Stripped("///@}}"),
    ]


def _generate_unwrap() -> Stripped:
    """Generate the main unwrapping function."""
    function_name = cpp_naming.function_name(Identifier("unwrap"))

    body = Stripped(
        f"""\
const std::shared_ptr<impl::IEnhanced<E> >& maybe_enhanced(
{I}std::dynamic_pointer_cast<impl::IEnhanced<E> >(that)
);

if (!maybe_enhanced) {{
{I}return nullptr;
}}

return maybe_enhanced->enhancement();"""
    )

    return Stripped(
        f"""\
/**
 * Try to unwrap the enhancement from \\p that instance.
 *
 * \\param that instance possibly wrapped with an enhancement
 * \\return the enhancement, or `nullptr` if \\p that instance has not been wrapped
 * \\tparam E type of the enhancement
 */
template <typename E>
std::shared_ptr<E> {function_name}(
{I}const std::shared_ptr<types::IClass>& that
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_must_unwrap() -> Stripped:
    """Generate the main unwrapping function which throws if no enhancement present."""
    unwrap_name = cpp_naming.function_name(Identifier("unwrap"))
    function_name = cpp_naming.function_name(Identifier("must_unwrap"))

    enhancement_var = cpp_naming.variable_name(Identifier("enhancement"))

    model_type_getter = cpp_naming.getter_name(Identifier("model_type"))

    body = Stripped(
        f"""\
std::shared_ptr<E> {enhancement_var}(
{I}{unwrap_name}<E>(that)
);
if (!{enhancement_var}) {{
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Expected an instance of ",
{III}stringification::to_string(that->{model_type_getter}()),
{III}" to have been already wrapped with an enhancement, "
{III}"but it has been not."
{II})
{I});
}}
return {enhancement_var};"""
    )

    return Stripped(
        f"""\
/**
 * Unwrap the enhancement from \\p that instance.
 *
 * \\remark \\p that instance must have been wrapped before.
 *
 * \\param that instance expected to be wrapped with an enhancement
 * \\return the enhancement
 * \\throw std::invalid_argument if \\p that instance has not been wrapped
 * \\tparam E type of the enhancement
 */
template <typename E>
std::shared_ptr<E> {function_name}(
{I}const std::shared_ptr<types::IClass>& that
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ code for wrapping model classes with custom enhancements."""
    namespace = Stripped(f"{library_namespace}::enhancing")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/stringification.hpp"
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <sstream>
#include <stdexcept>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            """\
/**
 * \\defgroup enhancing Enhance instances of the model with your custom enhancements.
 * @{
 */
namespace enhancing {"""
        ),
        Stripped("// region Forward declarations"),
        *_generate_wrap_forward_declarations(symbol_table=symbol_table),
        Stripped("// endregion Forward declarations"),
        Stripped(
            """\
/// \\cond HIDDEN
namespace impl {"""
        ),
        _generate_enhanced_interface_definition(),
        *[_generate_enhanced_class(cls) for cls in symbol_table.concrete_classes],
        *[_generate_concrete_wrap(cls=cls) for cls in symbol_table.concrete_classes],
        Stripped(
            f"""\
/**
 * Assert that the \\p that instance has not been already enhanced.
 *
 * \\param that instance to be checked
 * \\tparam E type of the enhancement
 * \\tparam T interface type of \\p that instance
 * \\throw std::logic_error if \\p that already enhanced
 */
template<
{I}typename E,
{I}typename T,
{I}typename std::enable_if<
{II}std::is_base_of<types::IClass, T>::value
{I}>::type* = nullptr
>
void AssertNotEnhanced(
{I}const std::shared_ptr<T>& that
) {{
{I}std::shared_ptr<impl::IEnhanced<E> > enhanced(
{II}std::dynamic_pointer_cast<
{III}impl::IEnhanced<E>
{II}>(that)
{I});
{I}if (enhanced != nullptr) {{
{II}throw std::logic_error(
{III}common::Concat(
{IIII}"An instance of ",
{IIII}stringification::to_string(that->model_type()),
{IIII}" has been already wrapped."
{III})
{II});
{I}}}
}}"""
        ),
        Stripped(
            """\
}  // namespace impl
/// \\endcond"""
        ),
        *_generate_wrap(symbol_table),
        _generate_unwrap(),
        _generate_must_unwrap(),
        Stripped(
            """\
}  // namespace enhancing
/**@}*/"""
        ),
        cpp_common.generate_namespace_closing(library_namespace),
        cpp_common.WARNING,
        Stripped(f"#endif  // {include_guard_var}"),
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate the unit test code for C#."""

from aas_core_codegen.csharp.tests import (
    _generate_common,
    _generate_common_json,
    _generate_common_jsonization,
    _generate_test_copying,
    _generate_test_descend_and_visitor_through,
    _generate_test_descend_once,
    _generate_test_enhancing,
    _generate_test_jsonization_of_concrete_classes,
    _generate_test_jsonization_of_enums,
    _generate_test_jsonization_of_interfaces,
    _generate_test_over_x_or_empty,
    _generate_test_verification_of_enums,
    _generate_test_x_or_default,
    _generate_test_xmlization_errors,
    _generate_test_xmlization_of_concrete_classes,
    _generate_test_xmlization_of_interfaces,
)

generate_common = _generate_common.generate
generate_common_json = _generate_common_json.generate
generate_common_jsonization = _generate_common_jsonization.generate
generate_test_copying = _generate_test_copying.generate
generate_test_descend_and_visitor_through = (
    _generate_test_descend_and_visitor_through.generate
)
generate_test_descend_once = _generate_test_descend_once.generate
generate_test_enhancing = _generate_test_enhancing.generate
generate_test_jsonization_of_concrete_classes = (
    _generate_test_jsonization_of_concrete_classes.generate
)
generate_test_jsonization_of_enums = _generate_test_jsonization_of_enums.generate
generate_test_jsonization_of_interfaces = (
    _generate_test_jsonization_of_interfaces.generate
)
generate_test_over_x_or_empty = _generate_test_over_x_or_empty.generate
generate_test_verification_of_enums = _generate_test_verification_of_enums.generate
generate_test_x_or_default = _generate_test_x_or_default.generate
generate_test_xmlization_errors = _generate_test_xmlization_errors.generate
generate_test_xmlization_of_concrete_classes = (
    _generate_test_xmlization_of_concrete_classes.generate
)
generate_test_xmlization_of_interfaces = (
    _generate_test_xmlization_of_interfaces.generate
)



"""Generate the names for the OPC UA nodeset definitions."""
import re

from aas_core_codegen.common import Identifier, Stripped
from aas_core_codegen import naming


def enum_name(identifier: Identifier, prefix: Identifier) -> Identifier:
    """
    Generate the name in OPC UA corresponding to the given enum identifier.

    >>> enum_name(Identifier("Something"), Identifier("AAS"))
    'AASSomethingDataType'

    >>> enum_name(Identifier("Something_better"), Identifier("AAS"))
    'AASSomethingBetterDataType'

    >>> enum_name(Identifier("Some_URL"), Identifier("AAS"))
    'AASSomeUrlDataType'
    """
    return Identifier(prefix + naming.capitalized_camel_case(identifier) + "DataType")


def constrained_primitive_name(
    identifier: Identifier, prefix: Identifier
) -> Identifier:
    """
    Generate the name in OPC UA for the given constrained primitive identifier.

    >>> constrained_primitive_name(Identifier("Something"), Identifier("AAS"))
    'AASSomethingDataType'

    >>> constrained_primitive_name(Identifier("Something_better"), Identifier("AAS"))
    'AASSomethingBetterDataType'

    >>> constrained_primitive_name(Identifier("Some_URL"), Identifier("AAS"))
    'AASSomeUrlDataType'
    """
    return Identifier(prefix + naming.capitalized_camel_case(identifier) + "DataType")


def class_name(identifier: Identifier, prefix: Identifier) -> Identifier:
    """
    Generate the name in OPC UA corresponding to the given class identifier.

    >>> class_name(Identifier("Something"), Identifier("AAS"))
    'AASSomethingType'

    >>> class_name(Identifier("Something_better"), Identifier("AAS"))
    'AASSomethingBetterType'

    >>> class_name(Identifier("Some_URL"), Identifier("AAS"))
    'AASSomeUrlType'
    """
    return Identifier(prefix + naming.capitalized_camel_case(identifier) + "Type")


def interface_name(identifier: Identifier, prefix: Identifier) -> Identifier:
    """
    Generate the interface name in OPC UA corresponding to the given identifier.

    >>> interface_name(Identifier("Something"), Identifier("AAS"))
    'IAASSomethingType'

    >>> interface_name(Identifier("Something_better"), Identifier("AAS"))
    'IAASSomethingBetterType'

    >>> interface_name(Identifier("Some_URL"), Identifier("AAS"))
    'IAASSomeUrlType'
    """
    return Identifier("I" + prefix + naming.capitalized_camel_case(identifier) + "Type")


def enum_literal_name(identifier: Identifier) -> Identifier:
    """
    Generate the name of an enumeration literal.

    >>> enum_literal_name(Identifier("Something"))
    'Something'

    >>> enum_literal_name(Identifier("Something_better"))
    'SomethingBetter'

    >>> enum_literal_name(Identifier("Some_URL"))
    'SomeUrl'
    """
    return naming.capitalized_camel_case(identifier)


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate the corresponding name for the property.

    >>> property_name(Identifier("Something"))
    'something'

    >>> property_name(Identifier("Something_better"))
    'somethingBetter'

    >>> property_name(Identifier("Some_URL"))
    'someUrl'
    """
    return naming.lower_camel_case(identifier)


_NON_ALPHANUMERIC = re.compile(r"[^A-Za-z0-9]+")


def constraint_browser_name(invariant_identifier: Stripped) -> Identifier:
    """
    Transform the invariant identifier into a valid browser name.

    >>> constraint_browser_name(Stripped('AASd-128'))
    'ConstraintAasd128'

    >>> constraint_browser_name(Stripped('AASc-3a-050'))
    'ConstraintAasc3a050'

    >>> constraint_browser_name(Stripped('1'))
    'Constraint1'
    """
    parts = _NON_ALPHANUMERIC.split(invariant_identifier)

    parts_joined = "_".join(parts)

    return naming.capitalized_camel_case(Identifier(f"Constraint_{parts_joined}"))



"""Generate the test code for the ``DescendOnce`` methods."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the ``DescendOnce`` methods.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = [
        Stripped(
            f"""\
private static void CompareOrRerecordTrace(
{I}IClass instance,
{I}string expectedPath)
{{
{I}var writer = new System.IO.StringWriter();
{I}foreach (var descendant in instance.DescendOnce())
{I}{{
{II}writer.WriteLine(Aas.Tests.Common.Trace(descendant));
{I}}}

{I}string got = writer.ToString();

{I}if (Aas.Tests.Common.RecordMode)
{I}{{
{II}string? parent = Path.GetDirectoryName(expectedPath);
{II}if (parent != null)
{II}{{
{III}if (!Directory.Exists(parent))
{III}{{
{IIII}Directory.CreateDirectory(parent);
{III}}}
{II}}}

{II}System.IO.File.WriteAllText(expectedPath, got);
{I}}}
{I}else
{I}{{
{II}if (!System.IO.File.Exists(expectedPath))
{II}{{
{III}throw new System.IO.FileNotFoundException(
{IIII}$"The file with the recorded trace does not exist: {{expectedPath}}; " +
{IIII}"maybe you want to set the environment " +
{IIII}$"variable {{Aas.Tests.Common.RecordModeEnvironmentVariableName}}?");
{II}}}

{II}string expected = System.IO.File.ReadAllText(expectedPath);
{II}Assert.AreEqual(
{III}expected.Replace("\\r\\n", "\\n"),
{III}got.Replace("\\r\\n", "\\n"),
{III}$"The expected trace from {{expectedPath}} does not match the actual one");
{I}}}
}}"""
        )
    ]  # type: List[str]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}()
{{
{I}Aas.{cls_name_csharp} instance = (
{II}Aas.Tests.CommonJsonization.LoadMaximal{cls_name_csharp}());

{I}CompareOrRerecordTrace(
{II}instance,
{II}Path.Combine(
{III}Aas.Tests.Common.TestDataDir,
{III}"DescendOnce",
{III}{csharp_common.string_literal(cls_name_json)},
{III}"maximal.json.trace"));
}}  // public void Test_{cls_name_csharp}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Directory = System.IO.Directory;
using Path = System.IO.Path;

using NUnit.Framework; // can't alias

namespace {namespace}.Tests
{{
{I}public class TestDescendOnce
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestDescendOnce
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the invariant verifiers from the intermediate representation."""
import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Set,
    Mapping,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations, naming
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
    wrap_text_into_lines,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
    description as csharp_description,
    transpilation as csharp_transpilation,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree, retree as parse_retree


# region Verify


def verify(
    spec_impls: specific_implementations.SpecificImplementations,
    verification_functions: Sequence[intermediate.Verification],
) -> Optional[List[str]]:
    """Verify all the implementation snippets related to verification."""
    errors = []  # type: List[str]

    expected_keys = []  # type: List[specific_implementations.ImplementationKey]

    for func in verification_functions:
        if isinstance(func, intermediate.ImplementationSpecificVerification):
            expected_keys.append(
                specific_implementations.ImplementationKey(
                    f"Verification/{func.name}.cs"
                ),
            )

    for key in expected_keys:
        if key not in spec_impls:
            errors.append(f"The implementation snippet is missing for: {key}")

    if len(errors) == 0:
        return None

    return errors


# endregion

# region Generate


class _PatternVerificationTranspiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a statement of a pattern verification into C#."""

    def __init__(self, defined_variables: Set[Identifier]) -> None:
        """
        Initialize with the given values.

        The ``defined_variables`` are shared between different statement
        transpilations. It is also mutated when assignments are transpiled. We need to
        keep track of variables so that we know when we have to define them, and when
        we can simply assign them a value, if they have been already defined.
        """
        self.defined_variables = defined_variables

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_joined_str_values(
        self, values: Sequence[Union[str, parse_tree.FormattedValue]]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """Transform the values of a joined string to a C# string literal."""
        if all(isinstance(value, str) for value in values):
            return (
                Stripped(csharp_common.string_literal("".join(values))),  # type: ignore
                None,
            )

        needs_interpolation = False

        parts = []  # type: List[str]
        for value in values:
            if isinstance(value, str):
                string_literal = csharp_common.string_literal(
                    value.replace("{", "{{").replace("}", "}}")
                )

                # We need to remove double-quotes since we are joining everything
                # ourselves later.

                assert string_literal.startswith('"') and string_literal.endswith('"')

                string_literal_wo_quotes = string_literal[1:-1]
                parts.append(string_literal_wo_quotes)

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self.transform(value.value)
                if error is not None:
                    return None, error
                assert code is not None

                assert (
                    "\n" not in code
                ), f"New-lines are not expected in formatted values, but got: {code}"

                needs_interpolation = True
                parts.append(f"{{{code}}}")
            else:
                assert_never(value)

        writer = io.StringIO()

        if needs_interpolation:
            writer.write('$"')
        else:
            writer.write('"')

        for part in parts:
            writer.write(part)

        writer.write('"')

        return Stripped(writer.getvalue()), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, str):
            # NOTE (mristin, 2022-06-11):
            # We assume that all the string constants are valid regular expressions.

            regex, parse_error = parse_retree.parse(values=[node.value])
            if parse_error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(
                    parse_error.cursor
                )

                return (
                    None,
                    Error(
                        node.original_node,
                        f"The string constant could not be parsed "
                        f"as a regular expression: \n"
                        f"{parse_error.message}\n"
                        f"{regex_line}\n"
                        f"{pointer_line}",
                    ),
                )

            assert regex is not None
            parse_retree.fix_for_utf16_regex_in_place(regex)

            # NOTE (mristin, 2022-06-11):
            # Strictly speaking, this is a joined string with a single value, a string
            # literal.
            return self._transform_joined_str_values(
                values=parse_retree.render(regex=regex)
            )
        else:
            raise AssertionError(f"Unexpected {node=}")

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return Stripped(csharp_naming.variable_name(node.identifier)), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        regex, parse_error = parse_retree.parse(values=node.values)
        if parse_error is not None:
            regex_line, pointer_line = parse_retree.render_pointer(parse_error.cursor)

            return (
                None,
                Error(
                    node.original_node,
                    f"The joined string could not be parsed "
                    f"as a regular expression: \n"
                    f"{parse_error.message}\n"
                    f"{regex_line}\n"
                    f"{pointer_line}",
                ),
            )

        assert regex is not None
        parse_retree.fix_for_utf16_regex_in_place(regex)

        return self._transform_joined_str_values(
            values=parse_retree.render(regex=regex)
        )

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        assert isinstance(node.target, parse_tree.Name)
        variable = csharp_naming.variable_name(node.target.identifier)
        code, error = self.transform(node.value)
        if error is not None:
            return None, error
        assert code is not None

        if node.target.identifier in self.defined_variables:
            return Stripped(f"{variable} = {code};"), None

        else:
            self.defined_variables.add(node.target.identifier)
            return Stripped(f"var {variable} = {code};"), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_pattern_verification(
    verification: intermediate.PatternVerification,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verification function that checks the regular expressions."""
    # NOTE (mristin, 2021-12-19):
    # We assume that we performed all the checks at the intermediate stage.

    construct_name = csharp_naming.private_method_name(
        Identifier(f"construct_{verification.name}")
    )

    blocks = []  # type: List[Stripped]

    # region Construct block

    writer = io.StringIO()
    writer.write(
        f"""\
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
[CodeAnalysis.SuppressMessageAttribute("ReSharper", "IdentifierTypo")]
[CodeAnalysis.SuppressMessage("ReSharper", "StringLiteralTypo")]
private static Regex {construct_name}()
{{
"""
    )

    defined_variables = set()  # type: Set[Identifier]
    transpiler = _PatternVerificationTranspiler(defined_variables=defined_variables)

    for i, stmt in enumerate(verification.parsed.body):
        if i == len(verification.parsed.body) - 1:
            break

        code, error = transpiler.transform(stmt)
        if error is not None:
            return None, error
        assert code is not None

        writer.write(textwrap.indent(code, I))
        writer.write("\n")

    if len(verification.parsed.body) >= 2:
        writer.write("\n")

    pattern_expr, error = transpiler.transform(verification.pattern_expr)
    if error is not None:
        return None, error
    assert pattern_expr is not None

    # A pragmatic heuristics for breaking lines
    if len(pattern_expr) < 50:
        writer.write(textwrap.indent(f"return new Regex({pattern_expr});\n", I))
    else:
        writer.write(textwrap.indent(f"return new Regex(\n{I}{pattern_expr});\n", I))

    writer.write("}")

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region Initialize the regex

    # NOTE (mristin, 2022-05-05):
    # We make this property look "public" since it is static and read-only.
    regex_name = csharp_naming.property_name(Identifier(f"regex_{verification.name}"))

    blocks.append(
        Stripped(f"private static readonly Regex {regex_name} = {construct_name}();")
    )

    assert len(verification.arguments) == 1
    assert isinstance(
        verification.arguments[0].type_annotation, intermediate.PrimitiveTypeAnnotation
    )
    # noinspection PyUnresolvedReferences
    assert (
        verification.arguments[0].type_annotation.a_type
        == intermediate.PrimitiveType.STR
    )

    arg_name = csharp_naming.argument_name(verification.arguments[0].name)

    writer = io.StringIO()
    if verification.description is not None:
        comment, comment_errors = csharp_description.generate_comment_for_signature(
            verification.description
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    method_name = csharp_naming.method_name(verification.name)

    writer.write(
        f"""\
public static bool {method_name}(string {arg_name})
{{
{I}return {regex_name}.IsMatch({arg_name});
}}"""
    )

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    return Stripped(writer.getvalue()), None


class _TranspilableVerificationTranspiler(csharp_transpilation.Transpiler):
    """Transpile the body of a :class:`.TranspilableVerification`."""

    # fmt: off
    @require(
        lambda environment, verification:
        all(
            environment.find(arg.name) is not None
            for arg in verification.arguments
        ),
        "All arguments defined in the environment"
    )
    # fmt: on
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
        verification: intermediate.TranspilableVerification,
    ) -> None:
        """Initialize with the given values."""
        csharp_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

        self._argument_name_set = frozenset(arg.name for arg in verification.arguments)

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(csharp_naming.variable_name(node.identifier)), None

        if node.identifier in self._argument_name_set:
            return Stripped(csharp_naming.variable_name(node.identifier)), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_as_prop = csharp_naming.property_name(node.identifier)
            return Stripped(f"Aas.Constants.{constant_as_prop}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(csharp_naming.method_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return Stripped(csharp_naming.enum_name(node.identifier)), None

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to C#. We could not find it neither in the constants, nor in "
            f"verification functions, nor as an enumeration. "
            f"If you expect this name to be transpilable, please contact "
            f"the developers.",
        )


def _transpile_transpilable_verification(
    verification: intermediate.TranspilableVerification,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Transpile a verification function."""
    # fmt: off
    type_inference, error = (
        intermediate_type_inference.infer_for_verification(
            verification=verification,
            base_environment=environment
        )
    )
    # fmt: on

    if error is not None:
        return None, error

    assert type_inference is not None

    transpiler = _TranspilableVerificationTranspiler(
        type_map=type_inference.type_map,
        environment=type_inference.environment_with_args,
        symbol_table=symbol_table,
        verification=verification,
    )

    body = []  # type: List[Stripped]
    for node in verification.parsed.body:
        stmt, error = transpiler.transform(node)
        if error is not None:
            return None, Error(
                verification.parsed.node,
                f"Failed to transpile the verification function {verification.name!r}",
                [error],
            )

        assert stmt is not None
        body.append(stmt)

    writer = io.StringIO()
    if verification.description is not None:
        comment, comment_errors = csharp_description.generate_comment_for_signature(
            verification.description
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    method_name = csharp_naming.method_name(verification.name)

    if verification.returns is None:
        return_type = "void"
    else:
        return_type = csharp_common.generate_type(type_annotation=verification.returns)

    arg_defs = []  # type: List[Stripped]
    for arg in verification.arguments:
        arg_type = csharp_common.generate_type(arg.type_annotation)
        arg_name = csharp_naming.argument_name(arg.name)
        arg_defs.append(Stripped(f"{arg_type} {arg_name}"))

    if len(arg_defs) == 0:
        writer.write(
            f"""\
public static {return_type} {method_name}()
{{"""
        )
    else:
        writer.write(
            f"""\
public static {return_type} {method_name}(
"""
        )

        for i, arg_def in enumerate(arg_defs):
            if i > 0:
                writer.write(",\n")
            writer.write(textwrap.indent(arg_def, I))

        writer.write("\n)\n{")

    for stmt in body:
        writer.write("\n")
        writer.write(textwrap.indent(stmt, I))

    if len(body) > 0:
        writer.write("\n")

    writer.write(f"}}  // public static {return_type} {method_name}")

    return Stripped(writer.getvalue()), None


class _InvariantTranspiler(csharp_transpilation.Transpiler):
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
    ) -> None:
        """Initialize with the given values."""
        csharp_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(csharp_naming.variable_name(node.identifier)), None

        if node.identifier == "self":
            # The ``that`` refers to the argument of the verification function.
            return Stripped("that"), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_as_prop = csharp_naming.property_name(node.identifier)
            return Stripped(f"Aas.Constants.{constant_as_prop}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(csharp_naming.method_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return Stripped(csharp_naming.enum_name(node.identifier)), None

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to C#. We could not find it "
            f"neither in the local variables, "
            f"nor in the global constants, "
            f"nor in verification functions, "
            f"nor as an enumeration. "
            f"If you expect this name to be transpilable, please contact "
            f"the developers.",
        )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_invariant(
    invariant: intermediate.Invariant,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Translate the invariant from the meta-model into C# code."""
    # fmt: off
    type_map, inference_error = (
        intermediate_type_inference.infer_for_invariant(
            invariant=invariant,
            environment=environment
        )
    )
    # fmt: on

    if inference_error is not None:
        return None, inference_error

    assert type_map is not None

    transpiler = _InvariantTranspiler(
        type_map=type_map,
        environment=environment,
        symbol_table=symbol_table,
    )

    expr, error = transpiler.transform(invariant.parsed.body)
    if error is not None:
        return None, error

    assert expr is not None

    writer = io.StringIO()
    if len(expr) > 50 or "\n" in expr:
        writer.write("if (!(\n")
        writer.write(textwrap.indent(expr, I))
        writer.write("))\n{\n")
    else:
        no_parenthesis_type_in_this_context = (
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )

        if isinstance(invariant.parsed.body, no_parenthesis_type_in_this_context):
            not_expr = f"!{expr}"
        else:
            not_expr = f"!({expr})"

        writer.write(f"if ({not_expr})\n{{\n")

    writer.write(
        textwrap.indent(
            f"""\
yield return new Reporting.Error(
{I}"Invariant violated:\\n" +
""",
            I,
        )
    )

    # NOTE (mristin, 2022-04-08):
    # We need to wrap the description in multiple literals as a single long
    # string literal is often too much for the readability.
    invariant_description_lines = wrap_text_into_lines(invariant.description)

    for i, literal in enumerate(invariant_description_lines):
        if i < len(invariant_description_lines) - 1:
            writer.write(f"{II}{csharp_common.string_literal(literal)} +\n")
        else:
            writer.write(f"{II}{csharp_common.string_literal(literal)});")

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_enum_value_sets(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate a class that pre-computes the sets of allowed enumeration literals."""
    blocks = []  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        enum_name = csharp_naming.enum_name(enum.name)

        if len(enum.literals) == 0:
            blocks.append(
                Stripped(
                    f"""\
internal static readonly HashSet<int> For{enum_name} = new HashSet<int>();"""
                )
            )
        else:
            hash_set_writer = io.StringIO()
            hash_set_writer.write(
                f"""\
internal static readonly HashSet<int> For{enum_name} = new HashSet<int>\n{{\n
"""
            )

            for i, literal in enumerate(enum.literals):
                literal_name = csharp_naming.enum_literal_name(literal.name)
                hash_set_writer.write(f"{I}(int)Aas.{enum_name}.{literal_name}")
                if i < len(enum.literals) - 1:
                    hash_set_writer.write(",\n")
                else:
                    hash_set_writer.write("\n")

            hash_set_writer.write("};")

            blocks.append(Stripped(hash_set_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Hash allowed enum values for efficient validation of enums.
/// </summary>
internal static class EnumValueSet
{
"""
    )
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal static class EnumValueSet")

    return Stripped(writer.getvalue())


def _generate_verify_method(our_type: intermediate.OurType) -> Stripped:
    """Generate the name of the ``Verification.Verify*`` method."""
    if isinstance(our_type, intermediate.Enumeration):
        name = csharp_naming.enum_name(our_type.name)
        return Stripped(f"Verification.Verify{name}")

    elif isinstance(our_type, intermediate.ConstrainedPrimitive):
        name = csharp_naming.class_name(our_type.name)
        return Stripped(f"Verification.Verify{name}")

    elif isinstance(our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)):
        return Stripped("Verification.Verify")
    else:
        assert_never(our_type)

    raise AssertionError("Unexpected execution path")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_property(
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the snippet to transform a property to errors."""
    # NOTE (mristin, 2022-03-10):
    # Instead of writing here a complex but general solution with unrolling we choose
    # to provide a simple, but limited, solution. First, the meta-model is quite
    # limited itself at the moment, so the complexity of the general solution is not
    # warranted. Second, we hope that there will be fewer bugs in the simple solution
    # which is particularly important at this early adoption stage.
    #
    # We anticipate that in the future we will indeed need a general and complex
    # solution. Here are just some thoughts on how to approach it:
    # * Leave the pattern matching to produce more readable code for simple cases,
    # * Unroll only in case of composite types and optional composite types.

    type_anno = (
        prop.type_annotation
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        else prop.type_annotation.value
    )

    if isinstance(type_anno, intermediate.OptionalTypeAnnotation):
        return None, Error(
            prop.parsed.node,
            "We currently implemented verification based on a very limited "
            "pattern matching due to code simplicity. We did not handle "
            "the case of nested optional values. Please contact "
            "the developers if you need this functionality.",
        )
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        if isinstance(type_anno.items, intermediate.OptionalTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of optional values. Please contact "
                "the developers if you need this functionality.",
            )
        elif isinstance(type_anno.items, intermediate.ListTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of lists. Please contact "
                "the developers if you need this functionality.",
            )
        else:
            pass
    else:
        pass

    stmts = []  # type: List[Stripped]

    prop_name = csharp_naming.property_name(prop.name)
    prop_literal = csharp_common.string_literal(naming.json_property(prop.name))

    # NOTE (mristin, 2022-03-12):
    # For some unexplainable reason, C# compiler can not infer that properties which
    # are enumerations are not null after an ``if (that.someProperty != null)``.
    # Hence, we need to add a null-coalescing for these particular cases.
    # Otherwise, we can just stick to ``that.someProperty``.

    needs_null_coalescing = (
        isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        and isinstance(prop.type_annotation.value, intermediate.OurTypeAnnotation)
        and isinstance(prop.type_annotation.value.our_type, intermediate.Enumeration)
    )
    if needs_null_coalescing:
        source_expr = Stripped("value")
    else:
        source_expr = Stripped(f"that.{prop_name}")

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # There is nothing that we check for primitive types.
        return Stripped(""), None
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        verify_method = _generate_verify_method(our_type=type_anno.our_type)

        foreach_error_in_verify = (
            f"foreach (var error in {verify_method}({source_expr}))"
        )
        # Heuristic to break the lines, very rudimentary
        if len(foreach_error_in_verify) > 80:
            foreach_error_in_verify = textwrap.dedent(
                f"""\
                foreach (
                    {I}var error in {verify_method}(
                    {II}{source_expr}))"""
            )

        # We can't use textwrap.dedent due to foreach_snippet.
        stmts.append(
            Stripped(
                f"""\
{foreach_error_in_verify}
{{
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{prop_literal}));
{I}yield return error;
}}"""
            )
        )

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert not isinstance(
            type_anno.items,
            (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
        ), (
            "We chose to implement only a very limited pattern matching; "
            "see the note above in the code."
        )

        # NOTE (mristin, 2022-03-16):
        # We only descend into our classes here.
        if not isinstance(type_anno.items, intermediate.OurTypeAnnotation):
            return Stripped(""), None

        index_var = csharp_naming.variable_name(Identifier(f"index_{prop.name}"))
        verify_method = _generate_verify_method(type_anno.items.our_type)

        foreach_item_in_source_expr = f"foreach (var item in {source_expr})"
        # Rudimentary heuristics for line breaking
        if len(foreach_item_in_source_expr) > 80:
            foreach_item_in_source_expr = textwrap.dedent(
                f"""\
                foreach(
                {I}var item in {source_expr})"""
            )

        foreach_error_in_verify_item = f"foreach (var error in {verify_method}(item))"
        if len(foreach_error_in_verify_item) > 70:
            foreach_error_in_verify_item = textwrap.dedent(
                f"""\
                foreach (
                {I}var error in {verify_method}(item))"""
            )

        stmts.append(
            Stripped(
                f"""\
int {index_var} = 0;
{foreach_item_in_source_expr}
{{
{I}{indent_but_first_line(foreach_error_in_verify_item, I)}
{I}{{
{II}error.PrependSegment(
{III}new Reporting.IndexSegment(
{IIII}{index_var}));
{II}error.PrependSegment(
{III}new Reporting.NameSegment(
{IIII}{prop_literal}));
{II}yield return error;
{I}}}
{I}{index_var}++;
}}"""
            )
        )

    else:
        assert_never(type_anno)

    verify_block = Stripped("\n".join(stmts))
    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        if needs_null_coalescing:
            value_type = csharp_common.generate_type(prop.type_annotation.value)
            if isinstance(prop.type_annotation.value, intermediate.OurTypeAnnotation):
                our_type = prop.type_annotation.value.our_type
                if isinstance(
                    our_type,
                    (
                        intermediate.Enumeration,
                        intermediate.AbstractClass,
                        intermediate.ConcreteClass,
                    ),
                ):
                    value_type = Stripped(f"Aas.{value_type}")

            return (
                Stripped(
                    f"""\
if (that.{prop_name} != null)
{{
{I}// We need to help the static analyzer with a null coalescing.
{I}{value_type} value = that.{prop_name}
{II}?? throw new System.InvalidOperationException();
{I}{indent_but_first_line(verify_block, I)}
}}"""
                ),
                None,
            )

        else:
            return (
                Stripped(
                    f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(verify_block, I)}
}}"""
                ),
                None,
            )
    else:
        return verify_block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_for_class(
    cls: intermediate.ConcreteClass,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transform method to errors for the given concrete class."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    name = csharp_naming.class_name(cls.name)

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(our_type=cls),
    )

    for invariant in cls.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to transpile the invariant of the class {cls.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    for prop in cls.properties:
        block, error = _generate_transform_property(prop=prop)
        if error is not None:
            errors.append(error)
        else:
            assert block is not None
            if block != "":
                blocks.append(block)

    if len(errors) > 0:
        return None, errors

    if len(blocks) == 0:
        blocks.append(
            Stripped(
                f"""\
// No verification has been defined for {name}.
yield break;"""
            )
        )

    writer = io.StringIO()

    interface_name = csharp_naming.interface_name(cls.name)
    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    writer.write(
        f"""\
[CodeAnalysis.SuppressMessage("ReSharper", "NegativeEqualityExpression")]
public override IEnumerable<Reporting.Error> {transform_name}(
{I}Aas.{interface_name} that
)
{{
"""
    )

    for i, stmt in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(stmt, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_transformer(
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a transformer to double-dispatch an instance to errors."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            continue

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            continue

        elif isinstance(our_type, intermediate.AbstractClass):
            # The abstract classes are directly dispatched by the transformer,
            # so we do not need to handle them separately.
            pass

        elif isinstance(our_type, intermediate.ConcreteClass):
            if our_type.is_implementation_specific:
                transform_key = specific_implementations.ImplementationKey(
                    f"Verification/transform_{our_type.name}.cs"
                )

                implementation = spec_impls.get(transform_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The transformation snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {transform_key}",
                        )
                    )
                    continue

                blocks.append(spec_impls[transform_key])
            else:
                block, cls_errors = _generate_transform_for_class(
                    cls=our_type,
                    symbol_table=symbol_table,
                    base_environment=base_environment,
                )
                if cls_errors is not None:
                    errors.extend(cls_errors)
                else:
                    assert block is not None
                    blocks.append(block)
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
private class Transformer
{I}: Visitation.AbstractTransformer<IEnumerable<Reporting.Error>>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // private class Transformer")

    return Stripped(writer.getvalue()), None


def _generate_verify_enumeration(enumeration: intermediate.Enumeration) -> Stripped:
    """Generate the verify method to check that an enum is valid."""
    name = csharp_naming.enum_name(enumeration.name)

    return Stripped(
        f"""\
/// <summary>
/// Verify that <paramref name="that" /> is a valid enumeration value.
/// </summary>
public static IEnumerable<Reporting.Error> Verify{name}(
{I}Aas.{name} that)
{{
{I}if (!EnumValueSet.For{name}.Contains(
{II}(int)that))
{I}{{
{II}yield return new Reporting.Error(
{III}$"Invalid {name}: {{that}}");
{I}}}
}}"""
    )


def _generate_verify_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the verify function for the constrained primitives."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(
            our_type=constrained_primitive
        ),
    )

    for invariant in constrained_primitive.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to transpile the invariant of "
                    f"the constrained primitive {constrained_primitive.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    if len(blocks) == 0:
        blocks.append(
            Stripped(
                """\
// There is no verification specified.
yield break;"""
            )
        )

    # NOTE (mristin, 2022-03-16):
    # Constrained primitives are not really classes, but we simply use the naming
    # for classes here since we need to pick *something*.
    name = csharp_naming.class_name(constrained_primitive.name)

    that_type = csharp_common.PRIMITIVE_TYPE_MAP[constrained_primitive.constrainee]

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Verify the constraints of <paramref name="that" />.
/// </summary>
public static IEnumerable<Reporting.Error> Verify{name} (
{I}{that_type} that)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    assert len(errors) == 0
    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code of the structures based on the symbol table.

    The ``namespace`` defines the AAS C# namespace.
    """
    using_directives = []  # type: List[Stripped]
    using_directives.extend(
        csharp_common.generate_using_aas_directive_if_necessary(namespace)
    )

    using_directives.append(
        Stripped(
            """\
using CodeAnalysis = System.Diagnostics.CodeAnalysis;
using Regex = System.Text.RegularExpressions.Regex;

using System.Collections.Generic;  // can't alias
using System.Linq;  // can't alias"""
        )
    )

    blocks = [
        csharp_common.WARNING,
        Stripped("\n".join(using_directives)),
    ]  # type: List[Stripped]

    verification_blocks = []  # type: List[Stripped]
    errors = []  # type: List[Error]

    base_environment = intermediate_type_inference.populate_base_environment(
        symbol_table=symbol_table
    )

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            implementation_key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.cs"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        None,
                        f"The snippet for the verification function "
                        f"{verification.name!r} is missing: {implementation_key}",
                    )
                )
            else:
                verification_blocks.append(implementation)

        elif isinstance(verification, intermediate.PatternVerification):
            implementation, error = _transpile_pattern_verification(
                verification=verification
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                verification_blocks.append(implementation)

        elif isinstance(verification, intermediate.TranspilableVerification):
            implementation, error = _transpile_transpilable_verification(
                verification=verification,
                symbol_table=symbol_table,
                environment=base_environment,
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                verification_blocks.append(implementation)

        else:
            assert_never(verification)

    verification_blocks.append(_generate_enum_value_sets(symbol_table=symbol_table))

    verification_blocks.append(
        Stripped(
            f"""\
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
private static readonly Verification.Transformer _transformer = (
{I}new Verification.Transformer());"""
        )
    )

    transformer_block, transformer_errors = _generate_transformer(
        symbol_table=symbol_table,
        base_environment=base_environment,
        spec_impls=spec_impls,
    )
    if transformer_errors is not None:
        errors.extend(transformer_errors)
    else:
        assert transformer_block is not None
        verification_blocks.append(transformer_block)

    verification_blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Verify the constraints of <paramref name="that" /> recursively.
/// </summary>
/// <param name="that">
/// The instance of the meta-model to be verified
/// </param>
public static IEnumerable<Reporting.Error> Verify(Aas.IClass that)
{{
{I}foreach (var error in _transformer.Transform(that))
{I}{{
{II}yield return error;
{I}}}
}}"""
        )
    )

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            verification_blocks.append(
                _generate_verify_enumeration(enumeration=our_type)
            )
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            (
                constrained_primitive_block,
                constrained_primitive_errors,
            ) = _generate_verify_constrained_primitive(
                constrained_primitive=our_type,
                symbol_table=symbol_table,
                base_environment=base_environment,
            )

            if constrained_primitive_errors is not None:
                errors.extend(constrained_primitive_errors)
            else:
                assert constrained_primitive_block is not None
                verification_blocks.append(constrained_primitive_block)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # We provide a general dispatch function.
            pass
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    verification_writer = io.StringIO()
    verification_writer.write(
        f"""\
namespace {namespace}
{{
{I}/// <summary>
{I}/// Verify that the instances of the meta-model satisfy the invariants.
{I}/// </summary>
"""
    )

    # region Write an example usage

    first_cls = (
        symbol_table.classes[0] if len(symbol_table.classes) > 0 else None
    )  # type: Optional[intermediate.ClassUnion]

    if first_cls is not None:
        cls_name: str

        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = csharp_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = csharp_naming.variable_name(Identifier("an_instance"))

        verification_writer.write(
            # We can not use textwrap.dedent since we indent everything including the
            # first line.
            f"""\
{I}/// <example>
{I}/// Here is an example how to verify an instance of {cls_name}:
{I}/// <code>
{I}/// var {an_instance_variable} = new Aas.{cls_name}(
{I}///     // ... some constructor arguments ...
{I}/// );
{I}/// foreach (var error in Verification.Verify({an_instance_variable}))
{I}/// {{
{I}/// {I}System.Console.Writeln(
{I}/// {II}$"{{error.Cause}} at: " +
{I}/// {II}Reporting.GenerateJsonPath(error.PathSegments));
{I}/// }}
{I}/// </code>
{I}/// </example>
"""
        )

    # endregion

    verification_writer.write(
        f"""\
{I}public static class Verification
{I}{{
"""
    )

    for i, verification_block in enumerate(verification_blocks):
        if i > 0:
            verification_writer.write("\n\n")

        verification_writer.write(textwrap.indent(verification_block, II))

    verification_writer.write(f"\n{I}}}  // public static class Verification")
    verification_writer.write(f"\n}}  // namespace {namespace}")

    blocks.append(Stripped(verification_writer.getvalue()))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Generate C++ visitors to iterate over instances."""

from aas_core_codegen.cpp.visitation import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""
Parse and operate on regular expressions defined over a possibly-formatted string.

Due to time constraints, we do not handle the whole of Python regular expression,
but just a subset. This subset was exactly enough to handle all the expressions of the
current meta-models. With newer meta-models, the subset is expected to grow.
"""

from aas_core_codegen.parse.retree import (
    _parse,
    _types,
    _stringify,
    _render,
    _visitor,
    _fix,
)

parse = _parse.parse
Error = _parse.Error
Cursor = _parse.Cursor
render_pointer = _parse.render_pointer

Node = _types.Node
Char = _types.Char
Range = _types.Range
Concatenation = _types.Concatenation
SymbolKind = _types.SymbolKind
Symbol = _types.Symbol
Group = _types.Group
CharSet = _types.CharSet
Quantifier = _types.Quantifier
Term = _types.Term
UnionExpr = _types.UnionExpr
Regex = _types.Regex
TermValueUnion = _types.TermValueUnion
Visitor = _types.Visitor
Transformer = _types.Transformer

PassThroughVisitor = _visitor.PassThroughVisitor

dump = _stringify.dump

Renderer = _render.Renderer
render = _render.render

fix_for_utf16_regex_in_place = _fix.fix_for_utf16_regex_in_place



"""Generate C++ code to de/stringify enumerations and primitives."""

import io
from typing import List

from icontract import ensure

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_model_type_from_string_definition() -> List[Stripped]:
    """Generate the definition of translation of a model type from string."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))
    from_string = cpp_naming.function_name(Identifier("model_type_from_string"))
    must_from_string = cpp_naming.function_name(
        Identifier("must_model_type_from_string")
    )

    return [
        Stripped(
            f"""\
/**
 * Try to parse the \\p text as a model type literal.
 *
 * \\param text to be parsed
 * \\return literal, or nothing, if \\p text invalid
 */
common::optional<types::{enum_name}> {from_string}(
{I}const std::string& text
);"""
        ),
        Stripped(
            f"""\
/**
 * Parse the \\p text as a model type literal.
 *
 * \\param text to be parsed
 * \\return literal
 * \\throw std::invalid_argument if \\p text invalid
 */
types::{enum_name} {must_from_string}(
{I}const std::string& text
);"""
        ),
    ]


def _generate_model_type_from_string_implementation(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the implementation of translation of a model type from string."""
    blocks = []  # type: List[Stripped]

    # region Map

    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    items = []  # type: List[str]
    for cls in symbol_table.concrete_classes:
        literal_name = cpp_naming.enum_literal_name(cls.name)

        literal_value = cpp_common.string_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        items.append(
            Stripped(
                f"""\
{{
{I}{literal_value},
{I}types::{enum_name}::{literal_name}
}}"""
            )
        )

    map_name = cpp_naming.constant_name(Identifier("model_type_from_string_map"))

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const std::unordered_map<
{I}std::string,
{I}types::{enum_name}
> {map_name} = {{
{I}{indent_but_first_line(items_joined, I)}
}};"""
        )
    )

    # endregion

    # region From function

    from_string = cpp_naming.function_name(Identifier("model_type_from_string"))

    blocks.append(
        Stripped(
            f"""\
common::optional<types::{enum_name}> {from_string}(
{I}const std::string& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}return {{}};
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    # region Must-from function

    must_from_string = cpp_naming.function_name(
        Identifier("must_model_type_from_string")
    )

    blocks.append(
        Stripped(
            f"""\
types::{enum_name} {must_from_string}(
{I}const std::string& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}throw std::invalid_argument(
{III}common::Concat(
{IIII}"Unexpected {enum_name} literal: ",
{IIII}text
{III})
{II});
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    return blocks


def _generate_model_type_to_string_definition() -> Stripped:
    """Generate the definition for translating a string to a model type."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    model_type_arg = cpp_naming.argument_name(Identifier("model_type"))

    # NOTE (mristin, 2023-07-12):
    # We use ``to_string`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
/**
 * Translate the enumeration literal \\p {model_type_arg} to text.
 *
 * \\param {model_type_arg} to be converted into text
 * \\return text representation of \\p {model_type_arg}
 * \\throw std::invalid_argument if \\p {model_type_arg} invalid
 */
std::string to_string(
{I}types::{enum_name} {model_type_arg}
);"""
    )


def _generate_model_type_to_string_implementation(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the implementation for translating a string to a model type."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        literal_name = cpp_naming.enum_literal_name(cls.name)
        literal_value = cpp_common.string_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        case_blocks.append(
            Stripped(
                f"""\
case types::{enum_name}::{literal_name}:
return {literal_value};"""
            )
        )

    model_type_arg = cpp_naming.argument_name(Identifier("model_type"))

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected model type: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(
{IIIII}{model_type_arg}
{IIII})
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    # NOTE (mristin, 2023-07-12):
    # We use ``to_string`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
std::string to_string(
{I}types::{enum_name} {model_type_arg}
) {{
{I}switch ({model_type_arg}) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )


def _generate_enum_from_string_definition(
    enum: intermediate.Enumeration,
) -> List[Stripped]:
    """Generate the definition of translation of an enum from string."""
    enum_name = cpp_naming.enum_name(enum.name)
    from_string = cpp_naming.function_name(Identifier(f"{enum.name}_from_string"))
    must_from_string = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_string")
    )

    return [
        Stripped(
            f"""\
/**
 * Try to parse the \\p text as a literal of
 * types::{enum_name}.
 *
 * \\param text to be parsed
 * \\return literal, or nothing, if \\p text invalid
 */
common::optional<types::{enum_name}> {from_string}(
{I}const std::string& text
);"""
        ),
        Stripped(
            f"""\
/**
 * Parse the \\p text as a literal of
 * types::{enum_name}.
 *
 * \\param text to be parsed
 * \\return literal
 * \\throw std::invalid_argument if \\p text invalid
 */
types::{enum_name} {must_from_string}(
{I}const std::string& text
);"""
        ),
    ]


def _generate_enum_from_string_implementation(
    enum: intermediate.Enumeration,
) -> List[Stripped]:
    """Generate the implementation of translation of an  enum from string."""
    blocks = []  # type: List[Stripped]

    # region Map

    enum_name = cpp_naming.enum_name(enum.name)

    items = []  # type: List[str]
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literal_value = cpp_common.string_literal(literal.value)

        items.append(
            Stripped(
                f"""\
{{
{I}{literal_value},
{I}types::{enum_name}::{literal_name}
}}"""
            )
        )

    map_name = cpp_naming.constant_name(Identifier(f"{enum.name}_from_string_map"))

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const std::unordered_map<
{I}std::string,
{I}types::{enum_name}
> {map_name} = {{
{I}{indent_but_first_line(items_joined, I)}
}};"""
        )
    )

    # endregion

    # region From function

    from_string = cpp_naming.function_name(Identifier(f"{enum.name}_from_string"))

    blocks.append(
        Stripped(
            f"""\
common::optional<types::{enum_name}> {from_string}(
{I}const std::string& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}return {{}};
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    # region Must-from function

    must_from_string = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_string")
    )

    blocks.append(
        Stripped(
            f"""\
types::{enum_name} {must_from_string}(
{I}const std::string& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}throw std::invalid_argument(
{III}common::Concat(
{IIII}"Unexpected {enum_name} literal: ",
{IIII}text
{III})
{II});
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    return blocks


def _generate_enum_to_string_definition(enum: intermediate.Enumeration) -> Stripped:
    """Generate the definition for translating a string to an enum literal."""
    enum_name = cpp_naming.enum_name(enum.name)

    literal_arg = cpp_naming.argument_name(Identifier("literal"))

    # NOTE (mristin, 2023-07-12):
    # We use ``to_string`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
/**
 * Translate the enumeration literal \\p {literal_arg}
 * of types::{enum_name} to text.
 *
 * \\param {literal_arg} to be converted into text
 * \\return text representation of \\p {literal_arg}
 * \\throw std::invalid_argument if \\p {literal_arg} invalid
 */
std::string to_string(
{I}types::{enum_name} {literal_arg}
);"""
    )


def _generate_enum_to_string_implementation(enum: intermediate.Enumeration) -> Stripped:
    """Generate the implementation for translating a string to an enum literal."""
    enum_name = cpp_naming.enum_name(enum.name)

    case_blocks = []  # type: List[Stripped]
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literal_value = cpp_common.string_literal(literal.value)

        case_blocks.append(
            Stripped(
                f"""\
case types::{enum_name}::{literal_name}:
{I}return {literal_value};"""
            )
        )

    literal_arg = cpp_naming.argument_name(Identifier("literal"))

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected literal: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(
{IIIII}{literal_arg}
{IIII})
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    # NOTE (mristin, 2023-07-12):
    # We use ``to_string`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
std::string to_string(
{I}types::{enum_name} {literal_arg}
) {{
{I}switch ({literal_arg}) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )


def _generate_base64_encode_definition() -> Stripped:
    """Generate the definition of the base64 encoding of bytes to a string."""
    function_name = cpp_naming.function_name(Identifier("base64_encode"))
    return Stripped(
        f"""\
/**
 * Encode the \\p bytes with base64 to a std::string.
 *
 * \\param bytes to be encoded
 * \\return base64-encoding of \\p bytes
 */
std::string {function_name}(
{I}const std::vector<std::uint8_t>& bytes
);"""
    )


def _generate_base64_encode_implementation() -> List[Stripped]:
    """Generate the implementation of the base64 encoding of bytes to a string."""
    char_base64_table = cpp_naming.constant_name(Identifier("char_base64_table"))
    char_base64_table_len = cpp_naming.constant_name(
        Identifier("char_base64_table_len")
    )

    function_name = cpp_naming.function_name(Identifier("base64_encode"))
    # noinspection SpellCheckingInspection
    return [
        Stripped(
            """\
// The following encoder has been adapted from Jouni Malinen <j@w1.fi> to work with
// std::string. The original source code is available at:
// https://web.mit.edu/freebsd/head/contrib/wpa/src/utils/base64.c
//
// See also the following StackOverflow question for a benchmark:
// https://stackoverflow.com/questions/342409/how-do-i-base64-encode-decode-in-c/41094722#41094722"""
        ),
        Stripped(
            f"""\
constexpr std::size_t {char_base64_table_len} = 65;
static const unsigned char {char_base64_table}[{char_base64_table_len}](
{I}"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
);"""
        ),
        Stripped(
            f"""\
std::string {function_name}(
{I}const std::vector<std::uint8_t>& bytes
) {{
{I}// See: https://cplusplus.com/reference/vector/vector/data/.
{I}// The data is guaranteed to be a continuous block in memory.
{I}const unsigned char* const src(
{II}bytes.data()
{I});

{I}const std::size_t len = bytes.size();

{I}// 3-byte blocks to 4-byte
{I}const std::size_t olen = 4 * ((len + 2) / 3);

{I}// Integer overflow?
{I}if (olen < len) {{
{II}throw std::invalid_argument(
{III}common::Concat(
{IIII}"The calculation of the output length overflowed. "
{IIII}"The length was: ",
{IIII}std::to_string(len),
{IIII}", but the output length was calculated as: ",
{IIII}std::to_string(olen)
{III})
{II});
{I}}}

{I}std::string out_string;
{I}out_string.resize(olen);

{I}unsigned char* out(
{II}reinterpret_cast<unsigned char*>(
{III}&out_string[0]
{II})
{I});

{I}const unsigned char* const end = src + len;

{I}const unsigned char* in = src;
{I}unsigned char* pos = out;

{I}while (end - in >= 3) {{
{II}*pos++ = {char_base64_table}[in[0] >> 2];
{II}*pos++ = {char_base64_table}[((in[0] & 0x03) << 4) | (in[1] >> 4)];
{II}*pos++ = {char_base64_table}[((in[1] & 0x0f) << 2) | (in[2] >> 6)];
{II}*pos++ = {char_base64_table}[in[2] & 0x3f];
{II}in += 3;
{I}}}

{I}if (end - in) {{
{II}*pos++ = {char_base64_table}[in[0] >> 2];

{II}if (end - in == 1) {{
{III}*pos++ = {char_base64_table}[(in[0] & 0x03) << 4];
{III}*pos++ = '=';
{II}}} else {{
{III}*pos++ = {char_base64_table}[
{IIII}((in[0] & 0x03) << 4) | (in[1] >> 4)
{III}];
{III}*pos++ = {char_base64_table}[(in[1] & 0x0f) << 2];
{II}}}
{II}*pos++ = '=';
{I}}}

{I}return out_string;
}}"""
        ),
    ]


def _generate_base64_decode_definition() -> Stripped:
    """Generate the definition of the base64 decoding of a string to bytes."""
    # NOTE (mristin, 2023-07-12):
    # We decode intentionally from ``std::string``, and
    function_name = cpp_naming.function_name(Identifier("base64_decode"))
    return Stripped(
        f"""\
/**
 * Decode the \\p the text with base64 to bytes.
 *
 * \\remark \\parblock
 * We intentionally decode from std::string and *not* from std::wstring as
 * the de/serialization libraries currently work only with UTF-8 encoded strings.
 * \\endparblock
 *
 * \\param text to be decoded
 * \\return decoded bytes, or error message, if any.
 */
common::expected<
{I}std::vector<std::uint8_t>,
{I}std::string
> {function_name}(
{I}const std::string& text
);"""
    )


def _generate_base64_decode_implementation() -> List[Stripped]:
    """Generate the implementation of the base64 decoding of a string to bytes."""
    function_name = cpp_naming.function_name(Identifier("base64_decode"))
    base64_lookup = cpp_naming.constant_name(Identifier("base64_lookup"))

    char_base64_table = cpp_naming.constant_name(Identifier("char_base64_table"))
    char_base64_table_len = cpp_naming.constant_name(
        Identifier("char_base64_table_len")
    )

    construct_base64_lookup = cpp_naming.function_name(
        Identifier("construct_base64_lookup")
    )

    return [
        Stripped(
            """\
// The following decoder is vaguely based on:
// https://github.com/danguer/blog-examples/blob/master/js/base64-binary.js,
// https://github.com/niklasvh/base64-arraybuffer/blob/master/src/index.ts and
// https://github.com/beatgammit/base64-js/blob/master/index.js."""
        ),
        Stripped(
            f"""\
std::vector<std::uint8_t> {construct_base64_lookup}() {{
{I}std::vector<std::uint8_t> lookup(256, 255);
{I}for (std::uint8_t i = 0; i < {char_base64_table_len}; ++i) {{
{II}lookup.at({char_base64_table}[i]) = i;
{I}}}
{I}return lookup;
}}
const std::vector<std::uint8_t> {base64_lookup} = {construct_base64_lookup}();"""
        ),
        Stripped(
            f"""\
common::expected<
{I}std::vector<std::uint8_t>,
{I}std::string
> {function_name}(
{I}const std::string& text
) {{
{I}if (text.empty()) {{
{II}return std::vector<std::uint8_t>();
{I}}}

{I}const std::size_t len = text.size();
{I}std::size_t len_wo_pad = len;

{I}// NOTE (mristin):
{I}// Some implementations forget the padding, so we try to be robust and check
{I}// for the padding manually.
{I}std::size_t bytes_length = (len * 3) / 4;
{I}if (text[len - 1] == '=') {{
{II}bytes_length--;
{II}len_wo_pad--;

{II}if (text[len - 2] == '=') {{
{III}bytes_length--;
{III}len_wo_pad--;
{II}}}
{I}}}

{I}std::vector<std::uint8_t> bytes(bytes_length);

{I}const std::size_t base64_lookup_len = {base64_lookup}.size();

{I}std::size_t pointer = 0;

{I}for (std::size_t i = 0; i < len; i += 4) {{
{II}// NOTE (mristin):
{II}// Admittedly, this is very verbose code, but we want to be efficient, so we
{II}// opted for performance over readability here.

{II}const unsigned char code0 = text[i];
{II}if (code0 >= base64_lookup_len) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i),
{IIII}": ",
{IIII}std::to_string(code0),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code0)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}
{II}const std::uint8_t encoded0 = {base64_lookup}[code0];
{II}if (encoded0 == 255) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i),
{IIII}": ",
{IIII}std::to_string(code0),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code0)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}

{II}const unsigned char code1 = text[i + 1];
{II}if (code1 >= base64_lookup_len) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 1),
{IIII}": ",
{IIII}std::to_string(code1),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code1)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}
{II}const std::uint8_t encoded1 = {base64_lookup}[code1];
{II}if (encoded1 == 255) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 1),
{IIII}": ",
{IIII}std::to_string(code1),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code1)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}

{II}// We map padding to 65, which is the value of "A".
{II}const unsigned char code2 = i + 2 < len_wo_pad ? text[i + 2] : 65;
{II}if (code2 >= base64_lookup_len) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 2),
{IIII}": ",
{IIII}std::to_string(code2),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code2)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}
{II}const std::uint8_t encoded2 = {base64_lookup}[code2];
{II}if (encoded2 == 255) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 2),
{IIII}": ",
{IIII}std::to_string(code2),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code2)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}

{II}// We map padding to 65, which is the value of 'A'.
{II}const unsigned char code3 = i + 3 < len_wo_pad ? text[i + 3] : 65;
{II}if (code3 >= base64_lookup_len) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 3),
{IIII}": ",
{IIII}std::to_string(code3),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code3)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}
{II}const std::uint8_t encoded3 = {base64_lookup}[code3];
{II}if (encoded3 == 255) {{
{III}std::string message = common::Concat(
{IIII}"Expected a valid character from base64-encoded string, "
{IIII}"but got at index ",
{IIII}std::to_string(i + 3),
{IIII}": ",
{IIII}std::to_string(code3),
{IIII}" (code: ",
{IIII}std::to_string(static_cast<int>(code3)),
{IIII}")"
{III});

{III}return common::make_unexpected(message);
{II}}}

{II}if (pointer >= bytes_length) {{
{III}break;
{II}}}
{II}bytes[pointer] = (encoded0 << 2) | (encoded1 >> 4);
{II}pointer++;

{II}if (pointer >= bytes_length) {{
{III}break;
{II}}}
{II}bytes[pointer] = ((encoded1 & 15) << 4) | (encoded2 >> 2);
{II}pointer++;

{II}if (pointer >= bytes_length) {{
{III}break;
{II}}}
{II}bytes[pointer] = ((encoded2 & 3) << 6) | (encoded3 & 63);
{II}pointer++;
{I}}}

{I}return bytes;
}}"""
        ),
    ]


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ header code for stringification."""
    namespace = Stripped(f"{library_namespace}::stringification")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <string>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            """\
/**
 * \\defgroup stringification Translate to strings, and, where applicable, from strings.
 * @{
 */
namespace stringification {"""
        ),
        *_generate_model_type_from_string_definition(),
        _generate_model_type_to_string_definition(),
    ]

    for enum in symbol_table.enumerations:
        blocks.extend(_generate_enum_from_string_definition(enum=enum))
        blocks.append(_generate_enum_to_string_definition(enum=enum))

    blocks.extend(
        [
            _generate_base64_encode_definition(),
            _generate_base64_decode_definition(),
            cpp_common.generate_namespace_closing(namespace),
            Stripped("/**@}*/"),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ code for stringification."""
    namespace = Stripped(f"{library_namespace}::stringification")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/stringification.hpp"

#pragma warning(push, 0)
#include <unordered_map>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        *_generate_model_type_from_string_implementation(symbol_table=symbol_table),
        _generate_model_type_to_string_implementation(symbol_table=symbol_table),
    ]  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        blocks.extend(_generate_enum_from_string_implementation(enum=enum))
        blocks.append(_generate_enum_to_string_implementation(enum=enum))

    blocks.extend(
        [
            *_generate_base64_encode_implementation(),
            *_generate_base64_decode_implementation(),
            Stripped(
                """\
}  // namespace stringification
/**@}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Generate the test code for the JSON de/serialization of interfaces."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II, INDENT3 as III


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the JSON de/serialization of interfaces.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            interface_name_csharp = csharp_naming.interface_name(cls.interface.name)
            implementer_cls_name_csharp = csharp_naming.class_name(implementer_cls.name)

            blocks.append(
                Stripped(
                    f"""\
[Test]
public void Test_round_trip_{interface_name_csharp}_from_{implementer_cls_name_csharp}()
{{
{I}var instance = Aas.Tests.CommonJsonization.LoadMaximal{implementer_cls_name_csharp}();

{I}var jsonObject = Aas.Jsonization.Serialize.ToJsonObject(instance);

{I}var anotherInstance = Aas.Jsonization.Deserialize.{interface_name_csharp}From(
{II}jsonObject);

{I}var anotherJsonObject = Aas.Jsonization.Serialize.ToJsonObject(
{II}anotherInstance);

{I}Aas.Tests.CommonJson.CheckJsonNodesEqual(
{II}jsonObject,
{II}anotherJsonObject,
{II}out Aas.Reporting.Error? error);

{I}if (error != null)
{I}{{
{II}Assert.Fail(
{III}"When we serialize the complete instance of {implementer_cls_name_csharp} " +
{III}"as {interface_name_csharp}, we get an error in the round trip: " +
{III}$"{{Reporting.GenerateJsonPath(error.PathSegments)}}: " +
{III}error.Cause
{II});
{I}}}
}}  // void Test_round_trip_{interface_name_csharp}_from_{implementer_cls_name_csharp}"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestJsonizationOfInterfaces
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestJsonizationOfInterfaces
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the visitor classes based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Identifier,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)

# region Generate


def _generate_ivisitor(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the visitor interface."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    blocks = []  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not visit
    # them.
    for cls in symbol_table.concrete_classes:
        # NOTE (empwilli, 2023-12-14):
        # Operate on interfaces instead of classes We operate on *interfaces*
        # instead of concrete classes to allow for custom extensions and
        # wrappers around our model classes.
        #
        # Originally, we used type overloading to dispatch the visit calls. After we
        # decided to support custom wrappers and enhancements to our classes, we had
        # to switch here to interfaces instead of concrete classes. The type
        # overloading does not work anymore in this setting, as descendants of
        # *concrete* classes would be wrongly dispatched. That is why we dispatch
        # explicitly, by having different visit method names instead of mere
        # type overloads.
        interface_name = java_naming.interface_name(cls.name)
        visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
void {visit_name}(
{I}{interface_name} that
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Define the interface for a visitor which visits the instances of the model.
 *
 * <p>When you use the visitor, please always call the main dispatching method
 * {{@link IVisitor#visit}}. You should most probably never call the {{@code visit*}}
 * methods directly. They are only made public so that model classes can access them.
 */
public interface IVisitor
{{
{I}void visit(IClass that);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_visitor_through(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the visitor that simply iterates over the instances."""
    blocks = []  # type: List[Stripped]

    # Abstract classes are modeled as interfaces in Java, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
public void {visit_name}(
{I}{interface_name} that
) {{
{I}// Just descend through, do nothing with {{@code that}}
{I}for (IClass something : that.descendOnce()) {{
{II}visit(something);
{I}}}
}}"""
            )
        )

    visitor_blocks = "\n\n".join(blocks)

    code = Stripped(
        f"""\
import {package}.types.model.*;
import {package}.visitation.IVisitor;

/**
 * Just descend through the instances without any action.
 *
 * <p>This class is meaningless for itself. However, it is a good base if you
 * want to descend through instances and apply actions only on a subset of
 * classes.
 */
public class VisitorThrough implements IVisitor {{
{I}public void visit(IClass that) {{
{II}that.accept(this);
{I}}}

{I}{indent_but_first_line(visitor_blocks, I)}
}}"""
    )

    return code


def _generate_abstract_visitor(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the visitor that performs double-dispatch."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.IVisitor;"),
    ]  # type: List[Stripped]

    blocks = [
        Stripped(
            f"""\
public void visit(IClass that)
{{
{I}that.accept(this);
}}"""
        )
    ]  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
public abstract void {visit_name}(
{I}{interface_name} that
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        """\
/**
 * Perform double-dispatch to visit the concrete instances.
 */
public abstract class AbstractVisitor implements IVisitor
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_ivisitor_with_context(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the interface for the visitor with context."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    blocks = []  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not visit
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
void {visit_name}(
{I}{interface_name} that,
{I}ContextT context
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Define the interface for a visitor which visits the instances of the model.
 *
 * <p>When you use the visitor, please always call the main dispatching method
 * {{@link IVisitorWithContext#visit}}. You should most probably never call the {{@code visit}}
 * methods directly. They are only made public so that model classes can access them.
 *
 * @param <ContextT> structure of the context
 */
public interface IVisitorWithContext<ContextT>
{{
{I}void visit(IClass that, ContextT context);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_visitor_with_context(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the visitor with context that performs double-dispatch."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.IVisitorWithContext;"),
    ]  # type: List[Stripped]

    blocks = [
        Stripped(
            f"""\
public void visit(IClass that, ContextT context)
{{
{I}that.accept(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not visit
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
public abstract void {visit_name}(
{I}{interface_name} that,
{I}ContextT context
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Perform double-dispatch to visit the concrete instances
 * with context.
 *
 * @param <ContextT> structure of the context
 */
public abstract class AbstractVisitorWithContext<ContextT>
{I}implements IVisitorWithContext<ContextT>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_itransformer(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the transformer interface."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    blocks = []  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
T {transform_name}(
{I}{interface_name} that
);"""
            )
        )
    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Define the interface for a transformer which transforms recursively
 * the instances into something else.
 *
 * <p>When you use the transformer, please always call the main dispatching method
 * {{@link ITransformer#transform}}. You should most probably never call the {{@code transform}}
 * methods directly. They are only made public so that model classes can access them.
 *
 * @param <T> type of the transformation result
 */
public interface ITransformer<T> {{
{I}T transform(IClass that);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the most abstract transformer that merely double-dispatches."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.ITransformer;"),
    ]  # type: List[Stripped]

    blocks = [
        Stripped(
            f"""\
public T transform(IClass that)
{{
{I}return that.transform(this);
}}"""
        )
    ]  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
public abstract T {transform_name}(
{I}{interface_name} that
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        """\
/**
 * Perform double-dispatch to transform recursively
 * the instances into something else.
 *
 * @param <T> type of the transformation result
 */
public abstract class AbstractTransformer<T> implements ITransformer<T>
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_itransformer_with_context(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the interface for the transformer with context."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    blocks = []  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
T {transform_name}(
{I}{interface_name} that,
{I}ContextT context
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Define the interface for a transformer which recursively transforms
 * the instances into something else while the context is passed along.
 *
 * <p>When you use the transformer, please always call the main dispatching method
 * {{@link ITransformerWithContext#transform}}. You should most probably never call the {{@code transform}}
 * methods directly. They are only made public so that model classes can access them.
 *
 * @param <ContextT> structure of the context
 * @param <T> type of the transformation result
 */
public interface ITransformerWithContext<ContextT, T> {{
{I}T transform(IClass that, ContextT context);
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_abstract_transformer_with_context(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Stripped:
    """Generate the most abstract transformer that merely double-dispatches."""
    java_imports = [
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.ITransformerWithContext;"),
    ]  # type: List[Stripped]

    blocks = [
        Stripped(
            f"""\
public T transform(IClass that, ContextT context)
{{
{I}return that.transform(this, context);
}}"""
        )
    ]  # type: List[Stripped]

    # Abstract classes have no particular implementation, so we do not transform
    # them.
    for cls in symbol_table.concrete_classes:
        interface_name = java_naming.interface_name(cls.name)
        transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
public abstract T {transform_name}(
{I}{interface_name} that,
{I}ContextT context
);"""
            )
        )

    writer = io.StringIO()

    for java_import in java_imports:
        writer.write(f"{java_import}\n")

    if len(java_imports) > 0:
        writer.write("\n")

    writer.write(
        f"""\
/**
 * Perform double-dispatch to transform recursively
 * the instances into something else.
 *
 * <p>When you use the transformer, please always call the main dispatching method
 * {{@link AbstractTransformerWithContext#transform}}. You should most probably never call the {{@code transform}}
 * methods directly. They are only made public so that model classes can access them.
 *
 * @param <ContextT> structure of the context
 * @param <T> type of the transformation result
 */
public abstract class AbstractTransformerWithContext<ContextT, T>
{I}implements ITransformerWithContext<ContextT, T>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_java_file(
    class_name: Stripped, code: Stripped, package: java_common.PackageIdentifier
) -> java_common.JavaFile:
    file_name = Stripped(f"{class_name}.java")

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.visitation;"),
        code,
        java_common.WARNING,
    ]  # type: List[Stripped]

    file_code = Stripped("\n\n".join(blocks))

    code_with_eof = f"{file_code}\n"

    return java_common.JavaFile(
        file_name,
        code_with_eof,
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, package: java_common.PackageIdentifier
) -> Tuple[Optional[List[java_common.JavaFile]], Optional[List[Error]]]:
    """
    Generate the java code of the visitors based on the intermediate representation

    The ``package`` defines the root Java package.
    """

    files = [
        _generate_java_file(
            Stripped("IVisitor"),
            _generate_ivisitor(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("VisitorThrough"),
            _generate_visitor_through(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("AbstractVisitor"),
            _generate_abstract_visitor(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("IVisitorWithContext"),
            _generate_ivisitor_with_context(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("AbstractVisitorWithContext"),
            _generate_abstract_visitor_with_context(
                symbol_table=symbol_table, package=package
            ),
            package,
        ),
        _generate_java_file(
            Stripped("ITransformer"),
            _generate_itransformer(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("AbstractTransformer"),
            _generate_abstract_transformer(symbol_table=symbol_table, package=package),
            package,
        ),
        _generate_java_file(
            Stripped("ITransformerWithContext"),
            _generate_itransformer_with_context(
                symbol_table=symbol_table, package=package
            ),
            package,
        ),
        _generate_java_file(
            Stripped("AbstractTransformerWithContext"),
            _generate_abstract_transformer_with_context(
                symbol_table=symbol_table, package=package
            ),
            package,
        ),
    ]  # type: List[java_common.JavaFile]

    return files, None


# endregion
